use std::rc::Rc;

use error_iter::ErrorIter as _;
use log::error;
use pixels::{Pixels, SurfaceTexture};
use player::HOSTED_DEVICE;
use viperid::{Device, VResult, SCREEN_H, SCREEN_W};
// https://github.com/parasyte/pixels/blob/main/examples/minimal-web/src/main.rs
use winit::{dpi::LogicalSize, event::Event, event_loop::{ControlFlow, EventLoop}, window::WindowBuilder};

fn main() {
    std::panic::set_hook(Box::new(console_error_panic_hook::hook));
    console_log::init_with_level(log::Level::Trace).expect("error initializing logger");
    wasm_bindgen_futures::spawn_local(run());
}

async fn run() {
    let wasm = include_bytes!("../../../example_project/build/game.wasm");
    let mut executor = 
        player::Executor::new(wasm)
        .expect("should have been able to create executor");

    host(HOSTED_DEVICE.with(|d| d.share()), move || {
        executor.update();
        executor.get_error()?;
        Ok(())
    }).await.expect("error while hosting");
}

async fn host(
    device: Device,
    mut update: impl FnMut() -> VResult<()> + 'static
) -> VResult<()> {
    let event_loop = EventLoop::new();

    let size = LogicalSize::new(SCREEN_W as f64, SCREEN_H as f64);
    let window = WindowBuilder::new()
        .with_title("viperid")
        .with_inner_size(size)
        .with_min_inner_size(size)
        .build(&event_loop)
        .expect("WindowBuilder error");
    let window = Rc::new(window);

    #[cfg(target_arch = "wasm32")]
    {
        use wasm_bindgen::JsCast;
        use winit::platform::web::WindowExtWebSys;

        // Retrieve current width and height dimensions of browser client window
        let get_window_size = || {
            let client_window = web_sys::window().unwrap();
            LogicalSize::new(
                client_window.inner_width().unwrap().as_f64().unwrap(),
                client_window.inner_height().unwrap().as_f64().unwrap(),
            )
        };

        let window = Rc::clone(&window);

        // Initialize winit window with current dimensions of browser client
        // window.set_inner_size(get_window_size());

        let client_window = web_sys::window().unwrap();

        // Attach winit canvas to body element
        web_sys::window()
            .and_then(|win| win.document())
            .and_then(|doc| doc.body())
            .and_then(|body| {
                body.append_child(&web_sys::Element::from(window.canvas()))
                    .ok()
            })
            .expect("couldn't append canvas to document body");

            /*
        // Listen for resize event on browser client. Adjust winit window dimensions
        // on event trigger
        let closure = wasm_bindgen::closure::Closure::wrap(Box::new(move |_e: web_sys::Event| {
            let size = get_window_size();
            window.set_inner_size(size)
        }) as Box<dyn FnMut(_)>);
        client_window
            .add_event_listener_with_callback("resize", closure.as_ref().unchecked_ref())
            .unwrap();
        closure.forget();
        */
    }

    // let mut input = WinitInputHelper::new();
    let mut pixels = {
        let window_size = window.inner_size();
        let surface_texture = 
            SurfaceTexture::new(window_size.width, window_size.height, window.as_ref());
        Pixels::new_async(SCREEN_W as u32, SCREEN_H as u32, surface_texture).await.expect("Pixels error")
    };

    event_loop.run(move |event, _, control_flow| {
        if let Event::RedrawRequested(_) = event {
            let frame = pixels.frame_mut();
            if let Err(err) = update() {
                log_anyhow_error("update", err);
                *control_flow = ControlFlow::Exit;
                return;
            }
            device.shared.screen.to_bitmap_slice(frame).expect("to_bitmap_slice error");
            if let Err(err) = pixels.render() {
                log_error("pixels.render", err);
                *control_flow = ControlFlow::Exit;
                return;
            }
        }

        window.request_redraw();
    })
}

fn log_error<E: std::error::Error + 'static>(method_name: &str, err: E) {
    error!("{method_name}() failed: {err}");
    for source in err.sources().skip(1) {
        error!("  Caused by: {source}");
    }
}


fn log_anyhow_error(method_name: &str, err: anyhow::Error) {
    error!("{method_name}() failed: {err}");
}
